@@ -18,6 +18,8 @@ import feedparser
 import html2text
 
 
+import threading
+
 # use_default_colors was added in python 2.4
 # if we don't have it, don't bother with colors
 if not hasattr(curses, "use_default_colors"):
@@ -45,7 +47,7 @@ timeoutsocket.setDefaultSocketTimeout(20)
 
 BROWSER = r"lynx %s"
 COLOR = {}
-VERSION = "0.9.3"
+VERSION = "0.9.4"
 RCDIR = "%s/.goodnews-%s" % (os.getenv("HOME"), VERSION)
 
 OS = sys.platform
@@ -202,6 +204,9 @@ def draw_status(stdscr, txt, delay=0):
 
 
 
+
+    
+
 def draw_header (stdscr, filterstring="", use_colors=USE_COLORS):
     LINES, COLS = stdscr.getmaxyx()
     #stdscr.attron (curses.A_STANDOUT)
@@ -1412,6 +1417,84 @@ def load_state_config(stdscr):
     #FIXME: read color config from file here
     
 
+
+
+class UpdateFeedThread(threading.Thread):
+
+    def __init__(self,tlock, stdscr, cur_ptr, first_ptr, draw_feeds, highlighted, keybindings):
+
+        threading.Thread.__init__(self)
+
+        self.stdscr = stdscr
+        self.cur_ptr = cur_ptr
+        self.first_ptr = first_ptr
+        self.draw_feeds = draw_feeds
+        self.tlock = tlock
+        self.highlighted = ''
+
+    def draw_status_tlock(self, stdscr, txt, delay=0):
+        self.tlock.acquire()            
+        draw_status(stdscr, txt, delay)
+        self.tlock.release()            
+        
+    def run(self):
+        try:
+            self.real_run()
+
+            draw_feeds(self.stdscr, self.first_ptr, self.highlighted)
+            draw_footer(self.stdscr, cur_cmd_grp, keybindings)
+            stdscr.refresh()
+
+        except:
+            pass
+
+        
+    def real_run(self):
+
+        if not self.cur_ptr:
+            return 0
+        LINES, COLS = stdscr.getmaxyx()
+
+        self.draw_status_tlock(stdscr, "Downloading " + self.cur_ptr.feedurl[:COLS], 0)
+
+        try:
+            self.cur_ptr.feed = urllib.urlopen(self.cur_ptr.feedurl)
+            self.cur_ptr.feed_data = self.cur_ptr.feed.read()
+            self.cur_ptr.feed = StringIO.StringIO(self.cur_ptr.feed_data)
+
+        except IOError, err:
+            self.draw_status_tlock(stdscr, "Downloading: "+err[1][1], 1)
+            return 2
+        except timeoutsocket.Timeout, err:
+            self.draw_status_tlock(stdscr,str(err), 1)
+            return 1
+
+        if not self.cur_ptr.title:
+            self.cur_ptr.title = self.cur_ptr.feedurl
+        if not self.cur_ptr.link:
+            self.cur_ptr.link = self.cur_ptr.feedurl
+
+        if not self.cur_ptr.feed:
+            return 0
+
+        self.draw_status_tlock(stdscr, "Parsing " + self.cur_ptr.feedurl[:COLS], 0)
+        xml = FeedParser(self.cur_ptr)
+        if self.cur_ptr.bozo:
+            self.draw_status_tlock(stdscr, "Warning: bozo XML! Maybe not a RSS feed?", 2)
+
+
+        #Don't mess up highlight
+        self.cur_ptr.highlighted = None
+
+        #if cur_ptr.feed:
+        #    del(cur_ptr.feed)
+
+        return 1
+
+
+
+    
+        
 # Update feed from server.
 def update_feed(stdscr, cur_ptr):
 
@@ -1456,6 +1539,8 @@ def update_feed(stdscr, cur_ptr):
     return 1
 
 
+
+
 def update_all_feeds(stdscr, first_ptr, draw_feeds=None, highlighted='', keybindings={}):
     for cur_ptr in tuple(first_ptr):
         if update_feed(stdscr, cur_ptr) == 2:
@@ -1466,22 +1551,39 @@ def update_all_feeds(stdscr, first_ptr, draw_feeds=None, highlighted='', keybind
             stdscr.refresh()
 
 
-errors = 0
-good = 0
+
+def update_all_feeds(stdscr, first_ptr, draw_feeds=None, highlighted='', keybindings={}):
+    tlock = threading.Lock()
+
+    threads = []
+
+    # before we begin, redraw the feeds without a highlight bar
+    draw_feeds(stdscr, first_ptr, highlighted='')
+
+    for cur_ptr in tuple(first_ptr):
+
+        threads.append(UpdateFeedThread(tlock, stdscr, cur_ptr, first_ptr, draw_feeds, highlighted, keybindings))
+        
+        threads[-1].setDaemon(0)
+        threads[-1].start()
+
+
+    # don't return until all threads are finished,
+    # or else we risk all kind of bad things :-) 
+    while threading.activeCount() > 1:
+        pass
+
+
+
 #/* Load feed from disk. And call update_feed if neccessary. */
 def load_feed(stdscr,cur_ptr):
-    global errors, good
-    
 
     hashme = hashurl(cur_ptr.feedurl)
     fname = RCDIR + "/cache/%s" % (hashme)
 
     try:
-        good += 1
         cache = open(fname)
-        #draw_status (stdscr, "good: " + str(good), 1);
     except IOError:
-        errors += 1        
         tmp = "Cache for %s is toast. Reloading from server..." % cur_ptr.feedurl
         draw_status (stdscr, tmp, 0);
         if not update_feed(stdscr, cur_ptr):
@@ -1517,6 +1619,72 @@ def load_feed(stdscr,cur_ptr):
     
     return 1
 
+
+class LoadFeedThread(threading.Thread):
+
+    def __init__(self, tlock, stdscr, cur_ptr):
+        threading.Thread.__init__(self)
+
+        self.tlock = tlock
+        self.stdscr = stdscr
+        self.cur_ptr = cur_ptr
+
+
+    def draw_status_tlock(self, stdscr, txt, delay=0):
+        self.tlock.acquire()            
+        draw_status(stdscr, txt, delay)
+        self.tlock.release()            
+        
+    def run(self):
+        try:
+            self.real_run()
+        except:
+            pass
+
+    def real_run(self):
+
+        tmp = "Loading cache for %s..." % self.cur_ptr.feedurl
+        self.draw_status_tlock(stdscr, tmp, 0)        
+
+        hashme = hashurl(self.cur_ptr.feedurl)
+        fname = RCDIR + "/cache/%s" % (hashme)
+
+        try:
+            cache = open(fname)
+        except IOError:
+            tmp = "Cache for %s is toast. Reloading from server..." % self.cur_ptr.feedurl
+            self.draw_status_tlock(self.stdscr, tmp, 0);
+            if not update_feed(self.stdscr, self.cur_ptr):
+                first_ptr.remove(self.cur_ptr)
+
+            return 0
+
+
+        #/* Read complete cachefile. */
+        self.cur_ptr.feed = cache
+        self.cur_ptr.feed_data = self.cur_ptr.feed.read()
+        self.cur_ptr.feed = StringIO.StringIO(self.cur_ptr.feed_data)    
+
+        #/* After loading FeedParserize the mess. */
+        xml = FeedParser(self.cur_ptr)
+        cache.close()
+
+        if self.cur_ptr.bozo:
+            self.draw_status_tlock(self.stdscr, "Warning: bozo XML! Maybe not a RSS feed?", 2)
+
+
+        # get readstatus
+        fname = RCDIR + "/cache-readstatus/%s" % (hashme)
+        cache_readstatus = open(fname)
+        for i in cache_readstatus.readlines():
+            id, status = i.split('\t')
+            if self.cur_ptr.items.has_key(id):
+                self.cur_ptr.items[id].readstatus = int(status)
+
+
+        return 1
+
+
 def load_all_feeds(stdscr):
     # work on a copy of first_ptr
     for cur_ptr in tuple(first_ptr):
@@ -1524,7 +1692,26 @@ def load_all_feeds(stdscr):
         draw_status(stdscr, tmp, 0)
         load_feed(stdscr, cur_ptr)
 
-    return 0
+
+
+"""
+# threaded version
+def load_all_feeds(stdscr):
+    tlock = threading.Lock()
+
+    threads = []
+
+    for cur_ptr in tuple(first_ptr):
+        threads.append(LoadFeedThread(tlock, stdscr, cur_ptr))
+        threads[-1].setDaemon(0)
+        threads[-1].start()
+
+    # don't return until all threads are finished,
+    # or else we risk all kind of bad things :-) 
+    while threading.activeCount() > 1:
+        pass
+
+"""
 
 
 
